!##############################################################################
!# 
!# Copyright by The HDF Group.
!# All rights reserved.
!#
!# This file is part of the hl_region High-Level HDF5 APIs.  The full copyright 
!# notice, including terms governing use, modification, and redistribution, 
!# is contained in the file COPYING, which can be found at the root of the
!# source code distribution tree and in the documentation directory (doc/html/). 
!# If you do not have access to this file, you may request a copy of 
!# "the hl_region High-Level HDF5 APIs copyright and license statement" from 
!# help@hdfgroup.org.
!# 
!##############################################################################

MODULE hl_region_H5LR

  USE hdf5
  USE ISO_C_BINDING

CONTAINS

  SUBROUTINE H5LRget_region_info_f(obj_id, ref, error, length, &               
       path, rank, dtype, sel_type, numelem, buf)

    IMPLICIT NONE

    INTEGER(hid_t), INTENT(IN) :: obj_id               ! Identifier of any object in an HDF5 file the region reference belongs to.
    TYPE(hdset_reg_ref_t_f), INTENT(IN), TARGET :: ref ! Region reference to query
    INTEGER, INTENT(OUT) :: error                      ! Error code:
                                                       !  0 on success and -1 on failure
                
    INTEGER(size_t), INTENT(INOUT), OPTIONAL :: length ! Size of the buffer to store path in
    CHARACTER*(*), INTENT(INOUT), OPTIONAL :: path     ! Full path that a region reference points to
    INTEGER, INTENT(OUT), OPTIONAL :: rank             ! The number of dimensions of the dataset dimensions 
                                                       !  of the dataset pointed by region reference
    INTEGER(hid_t), INTENT(OUT), OPTIONAL :: dtype     ! Datatype of the dataset pointed by the region reference
    INTEGER, INTENT(OUT), OPTIONAL :: sel_type         ! Type of the selection (point or hyperslab)
    INTEGER(size_t), INTENT(INOUT), OPTIONAL :: numelem ! Number of coordinate blocks or selected elements
    INTEGER(hsize_t), DIMENSION(*), OPTIONAL, TARGET :: buf ! Buffer containing description of the region pointed by region reference

    INTEGER(size_t) :: len_c
    TYPE(C_PTR) :: ref_c
    TYPE(C_PTR) :: buf_c
    INTEGER :: rank_c
    INTEGER(hid_t) :: dtype_c
    INTEGER :: sel_type_c
    INTEGER(size_t) :: numelem_c
    INTEGER(hsize_t), ALLOCATABLE, DIMENSION(:) :: buftmp

    INTEGER :: i,j,ii,itmp, path_size
    
    INTERFACE
       INTEGER FUNCTION H5LRget_region_info_c(obj_id, ref, length, &               
            path, path_size, rank, dtype, sel_type, numelem, buf) &
            bind(C, NAME='H5LRget_region_info_c')
         IMPORT
         INTEGER(hid_t)  :: obj_id
         TYPE(C_PTR), VALUE :: ref
         INTEGER(size_t) :: length
         CHARACTER(LEN=1) :: path
         INTEGER :: path_size
         INTEGER :: rank
         INTEGER(hid_t) :: dtype
         INTEGER :: sel_type
         INTEGER(size_t) :: numelem
         TYPE(C_PTR), VALUE :: buf
       END FUNCTION H5LRget_region_info_c
    END INTERFACE
    
    ref_c = C_LOC(ref)
    
    IF( PRESENT(length) ) THEN ! If option is present then either:
       IF(length .GT. 0 )THEN  !    - set the path length to the input value
          len_c = length       !
       ELSE                    ! or
          len_c = 0            !    - indicate to C routine length is unknown 
       ENDIF                   !      and should be returned
    ELSE
       len_c = -1              ! Else mark the optional parameter is absent
    ENDIF

    path_size = 0
    IF( PRESENT(path))THEN
       path_size = LEN(path)
    ENDIF

    rank_c = -1
    IF( PRESENT(rank)) rank_c = 0

    dtype_c = -1
    IF( PRESENT(dtype)) dtype_c = 0
     
    sel_type_c = -1
    IF( PRESENT(sel_type)) sel_type_c = 0
      
    IF( PRESENT(numelem) ) THEN ! If option is present then either:
       IF(numelem .GT. 0 )THEN  !    - set the numelem  to the input value
          numelem_c = numelem   !
       ELSE                     ! or
          numelem_c = 0         !    - indicate to C routine numelem is unknown 
       ENDIF                    !      and should be returned
    ELSE
       numelem_c = -1           ! Else mark the optional parameter is absent
    ENDIF

    IF( PRESENT(buf))THEN
       buf_c = C_LOC(buf)
    ELSE
       buf_c = C_NULL_PTR
    ENDIF

    error = H5LRget_region_info_c(obj_id, ref_c, len_c, &               
         path, path_size, rank_c, dtype_c, sel_type_c, numelem_c, buf_c)

    IF(PRESENT(rank)) rank = rank_c
    IF(PRESENT(numelem)) numelem = numelem_c
    IF(PRESENT(sel_type)) sel_type = sel_type_c
    IF(PRESENT(dtype)) dtype = dtype_c
    IF(PRESENT(length)) length = len_c

    IF(PRESENT(buf))THEN

       IF(sel_type_c.EQ.H5S_SEL_HYPERSLABS_F)THEN
          !
          ! Transpose dimension arrays because of C-FORTRAN storage order
          ! and subtract 1 to account for C starting arrays at 0.
          !
          ALLOCATE(buftmp(1:2*rank_c))
          ii = 1
          DO j = 1, 2
             DO i = 1, rank_c
                buftmp(j*rank_c-i+1) = buf(ii) + 1
                ii = ii + 1
             ENDDO
          ENDDO
          buf(1:2*rank_c) = buftmp(1:2*rank_c)
          DEALLOCATE(buftmp)
       ELSE IF(sel_type_c.EQ.H5S_SEL_POINTS_F)THEN
          DO i = 1, numelem
             buf(i) = buf(i) + 1
          ENDDO
       ENDIF
    ENDIF
       
  END SUBROUTINE H5LRget_region_info_f
  
  SUBROUTINE H5LRcreate_region_references_f(loc_id, num_elem, path, &
       block_coord, buf, error)
    
    IMPLICIT NONE
    
    INTEGER(hid_t), INTENT(IN)  :: loc_id                                       ! File identifier for the HDF5 file containing the 
                                                                                !  referenced regions or 
                                                                                !  an object identifier for any object in that file
    INTEGER(size_t), INTENT(IN) :: num_elem                                     ! Number of elements in the path and ref_buf arrays
    CHARACTER(LEN=*), DIMENSION(1:num_elem), INTENT(IN) :: path                 ! Array of strings, which contain the paths to the 
                                                                                !  target datasets for the region references
    INTEGER(hsize_t), DIMENSION(:), INTENT(IN) :: block_coord                   ! Array of hyperslab coordinates  
    TYPE(hdset_reg_ref_t_f), DIMENSION(1:num_elem), INTENT(OUT), TARGET :: buf  ! Buffer for returning an array of region references 
    INTEGER, INTENT(OUT) :: error                                               ! Error code:
                                                                                !  0 on success and -1 on failure
    TYPE(C_PTR) :: f_ptr
    
    INTEGER :: block_coord_size
    INTEGER, DIMENSION(1:num_elem) :: char_len
    INTEGER :: max_char_size
    INTEGER :: i
    INTEGER :: path_len

    INTERFACE
       INTEGER FUNCTION H5LRcreate_region_references_c(loc_id, num_elem, path, char_len, max_char_size, &
            block_coord_size, block_coord, f_ptr) bind(C, NAME='H5LRcreate_region_references_c')
         IMPORT
         INTEGER(hid_t)  :: loc_id
         INTEGER(size_t) :: num_elem
         INTEGER :: max_char_size
         INTEGER, DIMENSION(1:num_elem) :: char_len
         INTEGER :: block_coord_size
         INTEGER(hsize_t), DIMENSION(1:block_coord_size) :: block_coord
         CHARACTER(LEN=1), DIMENSION(1:num_elem) :: path
         TYPE(C_PTR), VALUE :: f_ptr
       END FUNCTION H5LRcreate_region_references_c
    END INTERFACE

    f_ptr = C_LOC(buf(1))

    ! Find the size of each character string in the array
    DO i = 1, num_elem
       char_len(i) = LEN_TRIM(path(i))
    ENDDO

    max_char_size = LEN(path(1))

    block_coord_size = SIZE(block_coord)

    error = H5LRcreate_region_references_c(loc_id, num_elem, path, char_len, max_char_size, &
         block_coord_size, block_coord, f_ptr)
    
  END SUBROUTINE H5LRcreate_region_references_f

  SUBROUTINE H5LRread_region_f(loc_id, ref, dtype, numelem, buf, error)

    IMPLICIT NONE
    
    INTEGER(hid_t), INTENT(IN) :: loc_id               ! File identifier for the HDF5 file containing the dataset with 
                                                       !  the referenced region or an object identifier for any object 
                                                       !  in that file 
    TYPE(hdset_reg_ref_t_f), INTENT(IN), TARGET :: ref ! Region reference specifying data to be read in 
    INTEGER(hid_t), INTENT(IN) :: dtype                ! Memory datatype of data read from referenced region into the 
                                                       !  application buffer
    INTEGER(size_t), INTENT(INOUT) :: numelem          ! Number of elements to be read into buffer buf
    TYPE(C_PTR), INTENT(OUT) :: buf                    ! Buffer in which data is returned to the application 
    INTEGER, INTENT(OUT) :: error                      ! Error code:
                                                       !  0 on success and -1 on failure
    TYPE(C_PTR) :: c_ref

    INTERFACE
       INTEGER FUNCTION H5LRread_region_c(loc_id, ref, dtype, numelem, buf) &
            bind(C, NAME='H5LRread_region_c')
         IMPORT
         INTEGER(hid_t) :: loc_id
         TYPE(C_PTR), VALUE :: ref
         INTEGER(hid_t) :: dtype
         INTEGER(size_t) :: numelem
         TYPE(C_PTR), VALUE :: buf
       END FUNCTION H5LRread_region_c
    END INTERFACE

    c_ref = C_LOC(ref)
    
    error = H5LRread_region_c(loc_id,  c_ref, dtype, numelem, buf)

  END SUBROUTINE H5LRread_region_f

! H5LRmake_dataset

  SUBROUTINE H5LRmake_dataset_f(loc_id, path, type_id, &
       buf_size,loc_id_ref, ref, error)

    IMPLICIT NONE

    INTEGER(hid_t), INTENT(IN) :: loc_id   ! Location identifier of the dataset to be created 
    CHARACTER(LEN=*), INTENT(IN) :: path   ! Path to the dataset being created 
    INTEGER(hid_t), INTENT(IN) :: type_id  ! Datatype of the dataset 
    INTEGER(size_t) :: buf_size            ! Size of the obj_id and buf arrays 
    INTEGER(hid_t), DIMENSION(1:buf_size), INTENT(IN), TARGET :: loc_id_ref ! Array of object identifiers; each identifier describes 
                                                                            ! to which HDF5 file the corresponding region reference belongs to 
    TYPE(hdset_reg_ref_t_f),  DIMENSION(1:buf_size), INTENT(IN), TARGET :: ref ! Array of region references  
    INTEGER, INTENT(OUT) :: error ! Error code:
                                  !  0 on success and -1 on failure

    TYPE(C_PTR) :: loc_id_ref_c, ref_c
    INTEGER(size_t) :: pathlen

    INTERFACE
       INTEGER FUNCTION H5LRmake_dataset_c(loc_id, pathlen, path, type_id, &
            buf_size,loc_id_ref_c, ref_c) &
            bind(C, NAME='H5LRmake_dataset_c')
         IMPORT

         INTEGER(hid_t) :: loc_id
         CHARACTER(LEN=1) :: path
         INTEGER(size_t) :: pathlen
         INTEGER(hid_t) :: type_id
         INTEGER(size_t) :: buf_size
         TYPE(C_PTR), VALUE :: loc_id_ref_c
         TYPE(C_PTR), VALUE :: ref_c
       END FUNCTION H5LRmake_dataset_c
    END INTERFACE

    ref_c = C_LOC(ref(1))
    loc_id_ref_c = C_LOC(loc_id_ref(1))

    pathlen = LEN_TRIM(path)

    error = H5LRmake_dataset_c(loc_id, pathlen, path, type_id, &
         buf_size,loc_id_ref_c, ref_c)


  END SUBROUTINE H5LRmake_dataset_f

  SUBROUTINE H5LRcopy_region_f(obj_id, ref, file_dest, path_dest, block_coord_dest, error)

    IMPLICIT NONE

    INTEGER(hid_t), INTENT(IN) :: obj_id                           ! Identifier of any object in a file 
                                                                   ! dataset region reference belongs to 
    TYPE(hdset_reg_ref_t_f), INTENT(IN), TARGET :: ref             ! Dataset region reference 
    CHARACTER(LEN=*), INTENT(IN) :: file_dest                      ! Name of the destination file 
    CHARACTER(LEN=*), INTENT(IN) :: path_dest                      ! Full path to the destination dataset
    INTEGER(hsize_t), DIMENSION(:), INTENT(IN) :: block_coord_dest ! Hyperslab coordinates in the destination dataset
    INTEGER, INTENT(OUT) :: error                                  ! Error code:
                                                                   !  0 on success and -1 on failure
    TYPE(C_PTR) :: ref_c
    INTEGER(size_t) :: file_dest_len, path_dest_len
    INTEGER :: nrank

    INTERFACE
       INTEGER FUNCTION H5LRcopy_region_c(obj_id, ref_c, file_dest_len, file_dest, path_dest_len, path_dest, &
            nrank, block_coord_dest) &
            bind(C, NAME='H5LRcopy_region_c')
         IMPORT

         INTEGER(hid_t) :: obj_id
         TYPE(C_PTR), VALUE :: ref_c 
         INTEGER(size_t) :: file_dest_len
         CHARACTER(LEN=1) :: file_dest
         INTEGER(size_t) :: path_dest_len
         CHARACTER(LEN=1) :: path_dest
         INTEGER :: nrank
         INTEGER(hsize_t), DIMENSION(1:2*nrank) :: block_coord_dest
       END FUNCTION H5LRcopy_region_c
    END INTERFACE

    ref_c = C_LOC(ref)

    nrank = SIZE(block_coord_dest)/2

    file_dest_len = LEN_TRIM(file_dest)
    path_dest_len = LEN_TRIM(path_dest)
    error = H5LRcopy_region_c(obj_id, ref_c, file_dest_len, file_dest, path_dest_len, path_dest, &
         nrank, block_coord_dest)


  END SUBROUTINE H5LRcopy_region_f

  
  SUBROUTINE H5LRcopy_reference_f( obj_id, ref, file, path, block_coord, ref_new, error )

    IMPLICIT NONE

    INTEGER(hid_t), INTENT(IN) :: obj_id                       ! Identifier of any object in a file an 
                                                               !  HDF5 reference belongs to 
    TYPE(hdset_reg_ref_t_f), INTENT(IN), TARGET :: ref         ! Reference to the datasets region 
    CHARACTER(LEN=*), INTENT(IN) :: file                       ! Name of the destination file
    CHARACTER(LEN=*), INTENT(IN) :: path                       ! Full path to the destination dataset 
    INTEGER(hsize_t), DIMENSION(:), INTENT(IN) :: block_coord  ! Hyperslab coordinates in the destination dataset 
    TYPE(hdset_reg_ref_t_f), INTENT(OUT), TARGET :: ref_new    ! Region reference to the new location of data 
    INTEGER, INTENT(OUT) :: error                              ! Error code:
                                                               !  0 on success and -1 on failure

    TYPE(C_PTR) :: ref_c, ref_new_c
    INTEGER(size_t) :: file_len, path_len
    INTEGER :: nrank

    INTERFACE
       INTEGER FUNCTION H5LRcopy_reference_c(obj_id, ref_c, file_len, file, path_len, path, &
            nrank, block_coord, ref_new_c ) &
            bind(C, NAME='H5LRcopy_reference_c')
         IMPORT

         INTEGER(hid_t) :: obj_id
         TYPE(C_PTR), VALUE :: ref_c 
         INTEGER(size_t) :: file_len
         CHARACTER(LEN=1) :: file
         INTEGER(size_t) :: path_len
         CHARACTER(LEN=1) :: path
         INTEGER :: nrank
         INTEGER(hsize_t), DIMENSION(1:2*nrank) :: block_coord
         TYPE(C_PTR), VALUE :: ref_new_c 
         
       END FUNCTION H5LRcopy_reference_c
    END INTERFACE

    ref_c = C_LOC(ref)
    ref_new_c = C_LOC(ref_new)

    nrank = SIZE(block_coord)/2

    file_len = LEN_TRIM(file)
    path_len = LEN_TRIM(path)

    error = H5LRcopy_reference_c(obj_id, ref_c, file_len, file, path_len, path, nrank, block_coord, ref_new_c )

  END SUBROUTINE H5LRcopy_reference_f


  SUBROUTINE H5LRcreate_ref_to_all_f(loc_id, group_path, ds_path, index_type, order, ref_type, error)
    
    IMPLICIT NONE

    INTEGER(hid_t), INTENT(IN) :: loc_id       ! File or group identifier 
    CHARACTER(LEN=*), INTENT(IN) :: group_path ! Absolute or relative path to the group at 
                                               !  which traversal starts 
    CHARACTER(LEN=*), INTENT(IN) :: ds_path    ! Absolute or relative path to the dataset 
                                               !  with region references to be created  
    INTEGER, INTENT(IN) :: index_type          ! Index type; valid values include:
                                               !     H5_INDEX_NAME_F
                                               !     H5_INDEX_CRT_ORDER_F
    INTEGER, INTENT(IN) :: order               ! Order in which index is traversed; valid values include:
                                               !     H5_ITER_DEC_F
                                               !     H5_ITER_INC_F
                                               !     H5_ITER_NATIVE_F
    INTEGER, INTENT(IN) :: ref_type            ! Reference type; valid values include:  
                                               !     H5R_DATASET_REGION_F
                                               !     H5R_OBJECT_F 
    INTEGER, INTENT(OUT) :: error              ! Error code:
                                               !  0 on success and -1 on failure

    INTEGER(size_t) :: group_path_len, ds_path_len

    INTERFACE
       INTEGER FUNCTION H5LRcreate_ref_to_all_c(loc_id, group_path_len, group_path, ds_path_len, &
            ds_path, index_type, order, ref_type) &
            bind(C, NAME='H5LRcreate_ref_to_all_c')
         IMPORT

         INTEGER(hid_t) :: loc_id
         INTEGER(size_t) :: group_path_len
         CHARACTER(LEN=1) :: group_path
         INTEGER(size_t) :: ds_path_len
         CHARACTER(LEN=1) :: ds_path
         INTEGER :: index_type, order, ref_type
         
       END FUNCTION H5LRcreate_ref_to_all_c
    END INTERFACE

    group_path_len = LEN_TRIM(group_path)
    ds_path_len = LEN_TRIM(ds_path)

    error= H5LRcreate_ref_to_all_c(loc_id, group_path_len, group_path, ds_path_len, &
         ds_path, index_type, order, ref_type)

  END SUBROUTINE H5LRcreate_ref_to_all_f

END MODULE hl_region_H5LR
