/*
 *****************************************************************************
 * PURPOSE
 *   This file contains C stubs for hl_region_H5LT Fortran APIs
 *
 * COPYRIGHT
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group.                                               *
 * Copyright by the Board of Trustees of the University of Illinois.         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of HDF5.  The full HDF5 copyright notice, including     *
 * terms governing use, modification, and redistribution, is contained in    *
 * the files COPYING and Copyright.html.  COPYING can be found at the root   *
 * of the source code distribution tree; Copyright.html can be found at the  *
 * root level of an installed copy of the electronic HDF5 document set and   *
 * is linked from the top-level documents page.  It can also be found at     *
 * http://hdfgroup.org/HDF5/doc/Copyright.html.  If you do not have          *
 * access to either file, you may request a copy from help@hdfgroup.org.     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 ******
*/
#include "hdf5.h"
#include "H5f90i_gen.h"
#include "H5f90i.h"
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include "hl_regionf90proto.h"
#include "../../src/h5hl_region.h"

/*-------------------------------------------------------------------------
 * Function: H5LTread_region_c
 *
 * Purpose: Calls H5LTread_region
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: December 28, 2009
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LTread_region_c(char *file, char *path, int_f *filelen, int_f *pathlen, 
		  const hsize_t_f *block_coord, int_f *rank, hid_t_f *mem_type, void *buf)
{

  herr_t status;
  int ret_value = -1;

  char *c_file = NULL;
  char *c_path = NULL;
  hsize_t *c_block_coord;
  int i,ii,j;
  float rdata[3][2];

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_file = (char *)HD5f2cstring(file, (size_t)*filelen)))
    return ret_value;
  if(NULL == (c_path = (char *)HD5f2cstring(path, (size_t)*pathlen)))
    return ret_value;

  c_block_coord = (hsize_t *)malloc(sizeof(hsize_t)*(*rank*2));
  if(!c_block_coord) return ret_value;

  /*
   * Transpose dimension arrays because of C-FORTRAN storage order
   * and subtract 1 to account for C starting arrays at 0.
   */
   ii = 0;
   for (j = 0; j < 2; j++) {
     for (i = 0; i < *rank ; i++) {
       c_block_coord[ii] =  block_coord[(j+1)*(*rank) - i -1] - 1;
       ii = ii + 1;
     }
   }

   /*
    * Call H5LTread_region
    */
   status = H5LTread_region(c_file, c_path, c_block_coord, (hid_t)*mem_type, buf);
   if ( status >= 0  ) ret_value = 0;
   
   if(c_file)
     free(c_file);
   if(c_path)
     free(c_path);
   
   free(c_block_coord);
   
   return ret_value;

}

/*-------------------------------------------------------------------------
 * Function: H5LTcopy_region_c
 *
 * Purpose: Calls H5LTcopy_region
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: December 28, 2009
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LTcopy_region_c(char *file_src,  int_f *file_src_len, char *path_src,  int_f *path_src_len, 
          const hsize_t_f *block_coord_src, int_f *block_src_rank, 
          char *file_dest,  int_f *file_dest_len, char *path_dest,  int_f *path_dest_len, 
          const hsize_t_f *block_coord_dest, int_f *block_dest_rank)
{

  herr_t status;
  int ret_value = -1;

  char *c_file_src  = NULL;
  char *c_path_src  = NULL;
  char *c_file_dest = NULL;
  char *c_path_dest = NULL;
  hsize_t *c_block_coord_src;
  hsize_t *c_block_coord_dest;
  int i,ii,j;

  /*
   * Convert FORTRAN name to C name
   */
  if(NULL == (c_file_src = (char *)HD5f2cstring(file_src, (size_t)*file_src_len)))
    return ret_value;
  if(NULL == (c_path_src = (char *)HD5f2cstring(path_src, (size_t)*path_src_len)))
    return ret_value;
  if(NULL == (c_file_dest = (char *)HD5f2cstring(file_dest, (size_t)*file_dest_len)))
    return ret_value;
  if(NULL == (c_path_dest = (char *)HD5f2cstring(path_dest, (size_t)*path_dest_len)))
    return ret_value;

  c_block_coord_src = (hsize_t *)malloc(sizeof(hsize_t)*(*block_src_rank*2));
  if(!c_block_coord_src) return ret_value;

  c_block_coord_dest = (hsize_t *)malloc(sizeof(hsize_t)*(*block_dest_rank*2));
  if(!c_block_coord_dest) return ret_value;

  /*
   * Transpose dimension arrays because of C-FORTRAN storage order
   * and subtract 1 to account for C starting arrays at 0.
   */
   ii = 0;
   for (j = 0; j < 2; j++) {
     for (i = 0; i < *block_src_rank ; i++) {
       c_block_coord_src[ii] =  block_coord_src[(j+1)*(*block_src_rank) - i -1] - 1;
       ii = ii + 1;
     }
   }
   ii = 0;
   for (j = 0; j < 2; j++) {
     for (i = 0; i < *block_dest_rank ; i++) {
       c_block_coord_dest[ii] =  block_coord_dest[(j+1)*(*block_dest_rank) - i -1] - 1;
       ii = ii + 1;
     }
   }

   /*
    * Call H5LTcopy_region
    */
   status = H5LTcopy_region(c_file_src, c_path_src, c_block_coord_src, 
			    c_file_dest, c_path_dest, c_block_coord_dest);
   if ( status >= 0  ) ret_value = 0;
   
   if(c_file_src)
     free(c_file_src);
   if(c_path_src)
     free(c_path_src);
   if(c_file_dest)
     free(c_file_dest);
   if(c_path_dest)
     free(c_path_dest);
   
   free(c_block_coord_src);
   free(c_block_coord_dest);
   
   return ret_value;
}

/*-------------------------------------------------------------------------
 * Function: H5LTread_bitfield_value_c
 *
 * Purpose: Calls H5LTread_bitfield_value
 *
 * Return: -1 on error, 0 on success
 *
 * Programmer: M. Scot Breitenfeld
 *
 * Date: December 28, 2009
 *
 * Comments:
 *
 * Modifications:
 *
 *-------------------------------------------------------------------------
 */
int_f
H5LTread_bitfield_value_c(hid_t_f *dset_id, int_f *num_values, int_f *offset,
			  int_f *lengths, hid_t_f *space, int_f *buf)
{
  herr_t status;
  int ret_value = -1;
  int i,j,k;

  unsigned *c_offset;
  unsigned *c_lengths;

  c_offset = (unsigned *)malloc(sizeof( unsigned)*(*num_values));
  if(!c_offset) return ret_value;
  
  c_lengths = (unsigned *)malloc(sizeof( unsigned)*(*num_values));
  if(!c_lengths) return ret_value;
  
  for (i=0; i<*num_values;i++) {
    c_offset[i]=(unsigned)offset[i]-1;
    c_lengths[i]=(unsigned)lengths[i];
  }

  /*
   * Call H5LTread_bitfield_value
   */
  status = H5LTread_bitfield_value((hid_t)*dset_id, (int)*num_values, c_offset, c_lengths, (hid_t)*space, (int*)buf );
  
  if ( status >= 0  ) ret_value = 0;
  
  return ret_value;
}
