!##############################################################################
!# 
!# Copyright by The HDF Group.
!# All rights reserved.
!#
!# This file is part of the hl_region High-Level HDF5 APIs.  The full copyright 
!# notice, including terms governing use, modification, and redistribution, 
!# is contained in the file COPYING, which can be found at the root of the
!# source code distribution tree and in the documentation directory (doc/html/). 
!# If you do not have access to this file, you may request a copy of 
!# "the hl_region High-Level HDF5 APIs copyright and license statement" from 
!# help@hdfgroup.org.
!# 
!##############################################################################

MODULE hl_region_H5LT

  USE hdf5
  USE ISO_C_BINDING

CONTAINS

  SUBROUTINE H5LTread_region_f(file, path, block_coord, mem_type, buf, error)

    IMPLICIT NONE
    
    CHARACTER(LEN=*), INTENT(IN) :: file                      ! Name of file
    CHARACTER(LEN=*), INTENT(IN) :: path                      ! Full path to a dataset 
    INTEGER(hsize_t), DIMENSION(:), INTENT(IN) :: block_coord ! Hyperslab coordinates 
    INTEGER(hid_t), INTENT(IN) :: mem_type                    ! Memory datatype, describing the buffer
                                                              ! the referenced data will be read into 
    TYPE(C_PTR), INTENT(OUT) :: buf                           ! Buffer containing data from 
                                                              ! the referenced region
    INTEGER, INTENT(OUT) :: error                             ! Error code:
                                                              ! 0 on success and -1 on failure
    INTEGER :: filelen, pathlen
    INTEGER :: rank
    
    INTERFACE
       INTEGER FUNCTION H5LTread_region_c(file, path, filelen, pathlen, block_coord, rank, mem_type, buf) &
            bind(C, NAME='H5LTread_region_c')
         IMPORT 
         CHARACTER(LEN=1) :: file, path
         INTEGER :: rank
         INTEGER(hsize_t), DIMENSION(1:rank) :: block_coord
         INTEGER(hid_t) :: mem_type 
         TYPE(C_PTR), VALUE :: buf
         INTEGER :: filelen, pathlen
       END FUNCTION H5LTread_region_c
    END INTERFACE
    
    filelen = LEN(file)
    pathlen = LEN(path)
    
    rank = SIZE(block_coord)/2

    error = H5LTread_region_c(file, path, filelen, pathlen, block_coord, rank, mem_type, buf)
    
  END SUBROUTINE H5LTread_region_f

  SUBROUTINE H5LTcopy_region_f(file_src, path_src, block_coord_src, &
       file_dest, path_dest, block_coord_dest, error)
    
    IMPLICIT NONE
    CHARACTER(LEN=*),INTENT(IN) :: file_src                       ! Name of the source file
    CHARACTER(LEN=*),INTENT(IN) :: path_src                       ! Full path to the source dataset 
    INTEGER(hsize_t), DIMENSION(:),INTENT(IN) :: block_coord_src  ! Hyperslab coordinates in the source dataset
    CHARACTER(LEN=*),INTENT(IN) :: file_dest                      ! Name of the destination file 
    CHARACTER(LEN=*),INTENT(IN) :: path_dest                      ! Full path to the destination dataset 
    INTEGER(hsize_t), DIMENSION(:),INTENT(IN) :: block_coord_dest ! Hyperslab coordinates in the destination dataset 
    INTEGER,INTENT(OUT) :: error                                  ! Error code:
                                                                  ! 0 on success and -1 on failure
    INTEGER :: block_dest_rank 
    INTEGER :: block_src_rank
    
    INTEGER :: file_src_len, path_src_len
    INTEGER :: file_dest_len, path_dest_len
    
    INTERFACE
       INTEGER FUNCTION H5LTcopy_region_c(file_src, file_src_len, path_src, path_src_len, &
            block_coord_src, block_src_rank, &
            file_dest, file_dest_len, path_dest, path_dest_len, &
            block_coord_dest, block_dest_rank) bind(C, NAME='H5LTcopy_region_c')
         
         IMPORT
         CHARACTER(LEN=1) :: file_src, path_src
         INTEGER :: file_src_len, path_src_len
         INTEGER :: block_src_rank
         INTEGER(hsize_t), DIMENSION(1:block_src_rank) :: block_coord_src
         CHARACTER(LEN=1) :: file_dest, path_dest
         INTEGER :: file_dest_len, path_dest_len
         INTEGER :: block_dest_rank
         INTEGER(hsize_t), DIMENSION(1:block_dest_rank) :: block_coord_dest
         
       END FUNCTION H5LTcopy_region_c
    END INTERFACE

    file_src_len = LEN(file_src)
    path_src_len = LEN(path_src)
    file_dest_len = LEN(file_dest)
    path_dest_len = LEN(path_dest)
    
    block_src_rank = SIZE(block_coord_src)/2
    block_dest_rank = SIZE(block_coord_dest)/2

    error = H5LTcopy_region_c(file_src, file_src_len, path_src, path_src_len, &
         block_coord_src, block_src_rank, &
         file_dest, file_dest_len, path_dest, path_dest_len, &
         block_coord_dest, block_dest_rank)

  END SUBROUTINE H5LTcopy_region_f

  SUBROUTINE H5LTread_bitfield_value_f(dset_id, num_values, offset, &
       lengths, space, buf, dims, error)
    IMPLICIT NONE
    INTEGER(hid_t), INTENT(IN) :: dset_id        ! Identifier of the dataset with bit-field values 
    INTEGER, INTENT(IN) :: num_values            ! Number of the values to be extracted 
    INTEGER, DIMENSION(*), INTENT(IN) :: offset  ! Array of staring bits to be extracted from the element
                                                 ! Valid values: 1 through 8
    INTEGER, DIMENSION(*), INTENT(IN) :: lengths ! Array of the number of bits to be extracted for each value 
    INTEGER, DIMENSION(:), INTENT(IN) :: dims    ! Dimension of the reading buffer, buf
    INTEGER(hid_t), INTENT(IN) :: space          ! Dataspace identifier, describing the elements to be read from the dataset 
                                                 ! with bit-field values 
    INTEGER, DIMENSION(*), INTENT(OUT), TARGET :: buf ! Buffer to read the values in, pass array with dimensions dims
    INTEGER, INTENT(OUT) :: error                ! Error code:
                                                 ! 0 on success and -1 on failure
    TYPE(C_PTR) :: f_ptr
    
    INTEGER :: rank,i,j,k,l,m,n,o, icnt, ik, ij, ii, ix
    INTEGER, DIMENSION(:), ALLOCATABLE :: buf_tmp
    INTEGER, DIMENSION(1:7) :: dims7
    LOGICAL, DIMENSION(1:7) :: mask
    INTEGER :: nsize
    
    INTERFACE
       INTEGER FUNCTION H5LTread_bitfield_value_c(dset_id, num_values, offset, &
            lengths, space, buf) bind(C, NAME='H5LTread_bitfield_value_c')
         
         IMPORT
         INTEGER(hid_t) :: dset_id
         INTEGER :: num_values
         INTEGER, DIMENSION(*) :: offset
         INTEGER, DIMENSION(*) :: lengths
         INTEGER(hid_t) :: space
         TYPE(C_PTR), VALUE :: buf
         
       END FUNCTION H5LTread_bitfield_value_c
    END INTERFACE
    
    f_ptr = C_LOC(buf(1))

    error = H5LTread_bitfield_value_c(dset_id, num_values, offset, &
         lengths, space, f_ptr)
    
    rank = SIZE(dims)

    dims7( :    ) = 0
    dims7(1:rank) = dims(1:rank)

    mask( :    ) = .FALSE.
    mask(1:rank) = .TRUE.

    nsize = PRODUCT(dims7,1,mask)

    ALLOCATE(buf_tmp(1:nsize))

    ! Transpose the array for the correct ordering
    ! Not done in place.
    
    IF(rank.EQ.1)THEN
       icnt = 1
       DO i = 1, dims7(1)
          ix = i
          buf_tmp(ix) = buf(icnt)
          icnt = icnt + 1
       ENDDO
    ELSE IF(rank.EQ.2)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             ix = i + &
                  dims7(1)*(j-1)
             buf_tmp(ix) = buf(icnt)
             icnt = icnt + 1
          ENDDO
       ENDDO
    ELSE IF(rank.EQ.3)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             DO k = 1, dims7(3)
                ix = j + &
                     dims7(2)*(i-1)+ &
                     dims7(1)*dims7(2)*(k-1)
                buf_tmp(ix) = buf(icnt)
                icnt = icnt + 1
             ENDDO
          ENDDO
       ENDDO
    ELSE IF(rank.EQ.4)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             DO k = 1, dims7(3)
                DO l = 1, dims7(4)
                   ix = k + &
                        dims7(3)*(j-1) + &
                        dims7(2)*dims7(3)*(i-1) + &
                        dims7(1)*dims7(2)*dims7(3)*(l-1)
                   buf_tmp(ix) = buf(icnt)
                   icnt = icnt + 1
                ENDDO
             ENDDO
          ENDDO
       ENDDO
    ELSE IF(rank.EQ.5)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             DO k = 1, dims7(3)
                DO l = 1, dims7(4)
                   DO m = 1, dims7(5)
                      ix = l + &
                           dims7(4)*(k-1) + &
                           dims7(3)*dims7(4)*(j-1) + &
                           dims7(2)*dims7(3)*dims7(4)*(i-1) + &
                           dims7(1)*dims7(2)*dims7(3)*dims7(4)*(m-1)
                      buf_tmp(ix) = buf(icnt)
                      icnt = icnt + 1
                   ENDDO
                ENDDO
             ENDDO
          ENDDO
       ENDDO
    ELSE IF(rank.EQ.6)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             DO k = 1, dims7(3)
                DO l = 1, dims7(4)
                   DO m = 1, dims7(5)
                      DO n = 1, dims7(6)
                         ix = m + &
                              dims7(5)*(l-1) + &
                              dims7(4)*dims7(5)*(k-1) + &
                              dims7(3)*dims7(4)*dims7(5)*(j-1) + &
                              dims7(2)*dims7(3)*dims7(4)*dims7(5)*(i-1) + &
                              dims7(1)*dims7(2)*dims7(3)*dims7(4)*dims7(5)*(n-1)
                         buf_tmp(ix) = buf(icnt)
                         icnt = icnt + 1
                      ENDDO
                   ENDDO
                ENDDO
             ENDDO
          ENDDO
       ENDDO
    ELSE IF(rank.EQ.7)THEN
       icnt = 1
       DO i = 1, dims7(1)
          DO j = 1, dims7(2)
             DO k = 1, dims7(3)
                DO l = 1, dims7(4)
                   DO m = 1, dims7(5)
                      DO n = 1, dims7(6)
                         DO o = 1, dims7(7)
                            ix = n + &
                                 dims7(6)*(m-1) + &
                                 dims7(5)*dims7(6)*(l-1) + &
                                 dims7(4)*dims7(5)*dims7(6)*(k-1) + &
                                 dims7(3)*dims7(4)*dims7(5)*dims7(6)*(j-1) + &
                                 dims7(2)*dims7(3)*dims7(4)*dims7(5)*dims7(6)*(i-1) + &
                                 dims7(1)*dims7(2)*dims7(3)*dims7(4)*dims7(5)*dims7(6)*(o-1)
                            buf_tmp(ix) = buf(icnt)
                            icnt = icnt + 1
                         ENDDO
                      ENDDO
                   ENDDO
                ENDDO
             ENDDO
          ENDDO
       ENDDO
    ENDIF

    buf(1:nsize) = buf_tmp(1:nsize)

    DEALLOCATE(buf_tmp)

  END SUBROUTINE H5LTread_bitfield_value_f

END MODULE hl_region_H5LT
